/* dib.cpp
*
*/
#define WIN32_LEAN_AND_MEAN
#include <stdlib.h>
#include <windows.h>
#include <malloc.h>
#include "Dib.h"

// Every DIB should start with 'BM'
#define DIB_MAGIC 0x4D42

#define MIN(a,b) ((a) < (b) ? (a) : (b))

/*
* size of scan in bytes =
*   Pixel Width * bits per pixel rounded up to a uint32 boundary
*/

inline long ScanBytes(int pixWidth, int bitsPixel) {
  return (((long)pixWidth*bitsPixel+31) / 32) * 4;
}

class RgbQuad : public tagRGBQUAD
{
public:
	RgbQuad(int r, int g, int b)
	{
		rgbRed = (BYTE)(WORD)r;
		rgbGreen = (BYTE)(WORD)g;
		rgbBlue = (BYTE)(WORD)b;
		rgbReserved = 0;
	}
    RgbQuad(const RGBQUAD far& q) {*(RGBQUAD*)this = q;}
};

const RgbQuad WinColors[16] = {
	RgbQuad(0,0,0),          // Black
	RgbQuad(128,0,0),        // DkRed
	RgbQuad(0,128,0),        // DkGreen
	RgbQuad(128,128,0),      // DkYellow
	RgbQuad(0,0,128),        // DkBlue
	RgbQuad(128,0,128),      // DkMagneta
	RgbQuad(0,128,128),      // DkCyan
	RgbQuad(128,128,128),    // Gray
	RgbQuad(192,192,192),    // LtGray
	RgbQuad(255,0,0),        // LtRed
	RgbQuad(0,255,0),        // LtGreen
	RgbQuad(255,255,0),      // LtYellow
	RgbQuad(0,0,255),        // LtBlue
	RgbQuad(255,0,255),      // LtMagneta
	RgbQuad(0,255,255),      // LtCyan
	RgbQuad(255,255,255)     // White
};

int NColors(WORD bitCount);

CDib::CDib()
{
	lpDib = 0;
	pBits = 0;
}

CDib::~CDib()
{
	Delete();
}


// Functions for extracting information from DIB memory blocks

int CDib::GetInfoHeaderSize ()
{

	return (lpDib? ((BITMAPINFOHEADER *) lpDib)->biSize: 0);
}

int CDib::GetWidth()
{
	return (lpDib? (int) (((BITMAPINFOHEADER *) lpDib)->biWidth): 0);
}

int CDib::GetHeight()
{
	return (lpDib? abs(((BITMAPINFOHEADER *) lpDib)->biHeight): 0);
}

int CDib::GetLineWidth()
{
	BITMAPINFOHEADER *pInfo = (BITMAPINFOHEADER *) lpDib;
	return ScanBytes(pInfo->biWidth,pInfo->biBitCount);
}

int CDib::GetBitsPerPixel()
{
	if (!IsValid()) return 0;
	BITMAPINFOHEADER *pInfo;
	pInfo = (BITMAPINFOHEADER *) lpDib;
	return pInfo->biBitCount;
}

RGBQUAD * CDib::GetColors()
{
	if (!IsValid()) return 0;
	return  (RGBQUAD *) (lpDib + GetInfoHeaderSize());
}


int CDib::GetNumColors()
{
	int BitCount;
	int NumColors;
	int HeaderSize;
	if (!IsValid()) return 0;
	BITMAPINFOHEADER *pInfo;
	pInfo = (BITMAPINFOHEADER *) lpDib;
	HeaderSize = GetInfoHeaderSize();
	BitCount = pInfo->biBitCount ;

	if (HeaderSize >= 36)
		NumColors = pInfo->biClrUsed ;
	else NumColors = 0 ;

	if (NumColors == 0) {
		if (BitCount < 24) NumColors = 1 << BitCount ;
	}
	return NumColors;
}
		

void CDib::CalcBits()
{
     int ColorTableSize ;
	 ColorTableSize = GetNumColors() * sizeof (RGBQUAD) ;	
     pBits = lpDib + GetInfoHeaderSize() + ColorTableSize;
}


//
// Construct a Dib given dimensions and color depth
//

BOOL CDib::Create(int width, int height, int nColorBits)
{
	if (lpDib) Delete();  // Release previously allocated memory
	
	int nColors = NColors(nColorBits);
	if (nColors<0) return FALSE;

    // Fill in the header info.
	BITMAPINFOHEADER      InfoHeader;
	InfoHeader.biSize       = sizeof(BITMAPINFOHEADER);
	InfoHeader.biWidth      = width;
	InfoHeader.biHeight     = height;
	InfoHeader.biBitCount   = nColorBits;
	InfoHeader.biPlanes     = 1;
	InfoHeader.biCompression   = BI_RGB;
	InfoHeader.biXPelsPerMeter = 0;
	InfoHeader.biYPelsPerMeter = 0;
	InfoHeader.biClrUsed       = 0; // use biBitCount to determine number of colors
	InfoHeader.biClrImportant  = 0; // all colors important
	InfoHeader.biSizeImage = ScanBytes(width, InfoHeader.biBitCount) * abs(height);

    // Allocate space for everything

	int  colorAlloc = nColors * sizeof(RGBQUAD); // size of color tables
	long bitsAlloc = InfoHeader.biSize + colorAlloc + InfoHeader.biSizeImage;

	lpDib = (BYTE *) GlobalAlloc(GMEM_FIXED|GMEM_ZEROINIT, bitsAlloc);
	if (!lpDib) return FALSE;

	BITMAPINFO *pInfo = GetInfo();

	pInfo->bmiHeader = InfoHeader;

	// Create a color table
	
	// Use WinColors (defined above)
	switch (nColors) {
	case 2:
		pInfo->bmiColors[0] = WinColors[0];
		pInfo->bmiColors[1] = WinColors[15];
		break;
	case 16:
		for (int i=0; i<16; i++) {
			pInfo->bmiColors[i] = WinColors[i];
		}
		break;
	}
	CalcBits();
	return TRUE;
}

// Read a DIB from a file into memory

int CDib::Read(char * szFileName)
{
     BITMAPFILEHEADER bmfh ;
     DWORD            dwDibSize, dwOffset, dwHeaderSize ;
     int              hFile ;
     WORD             wDibRead ;

     if (-1 == (hFile = _lopen (szFileName, OF_READ | OF_SHARE_DENY_WRITE)))
          return -1; // could not open file


     wDibRead = _lread (hFile, (LPSTR) &bmfh, sizeof (BITMAPFILEHEADER));

	if (bmfh.bfType != *(WORD *)"BM" || wDibRead != sizeof(BITMAPFILEHEADER)) {
		_lclose (hFile) ;
		return -2 ; // not a valid Windows bitmap file 
	}

	dwDibSize = bmfh.bfSize - sizeof (BITMAPFILEHEADER) ;

    lpDib = (BYTE * ) GlobalAlloc (GMEM_FIXED, dwDibSize) ;

	if (lpDib == 0) {
		_lclose (hFile) ;
		return -3 ; // failed to allocated memory
	}

	dwOffset = 0 ;
	while (dwDibSize > 0) {
		wDibRead = (WORD) MIN (32768ul, dwDibSize) ;

		if (wDibRead != _lread (hFile, (LPSTR) (lpDib + dwOffset), wDibRead)) {
			_lclose (hFile) ;
			Delete() ;
			return -4;  // unexpected end-of-file
		}
		dwDibSize -= wDibRead ;
		dwOffset  += wDibRead ;
	}
	_lclose (hFile) ;

	dwHeaderSize = GetInfoHeaderSize ();
	if (dwHeaderSize < sizeof(BITMAPINFOHEADER)) {
		Delete();
		return -5; // probably a Windows 1.x bitmap (BITMAPINFOCOREHEADER)
	}
	CalcBits();
	return 0;
}

int CDib::Write(char *szFileName)
{
	BITMAPFILEHEADER bmf;
	HANDLE  hFile ;
	DWORD  dwSize, dwDibWrite;
	BOOL bRet;

	if (!lpDib) return -1;
	dwSize = GlobalSize(lpDib);

	hFile = CreateFile(szFileName,
					   GENERIC_WRITE,
					   FILE_SHARE_READ,
					   NULL, // pointer to security attributes
					   CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
	if (!hFile) return -1; // could not create file

	// write file header

	bmf.bfType = DIB_MAGIC;
	bmf.bfSize = sizeof(bmf) + dwSize;
	bmf.bfReserved1 = 0;
	bmf.bfReserved2 = 0;
	bmf.bfOffBits = sizeof(bmf) + (int)(GetBits() - (BYTE *)GetInfo());
	bRet = WriteFile(hFile, &bmf, sizeof(bmf), &dwDibWrite, NULL);

	// write bitmapinfo

	bRet = WriteFile(hFile, lpDib, dwSize, &dwDibWrite, NULL);

	// close file
	
	CloseHandle(hFile);
	return 0;
}

void CDib::Delete()
{
	if (lpDib) {
		GlobalFree(lpDib);
		lpDib = 0;
	}
	pBits = 0;
}

void CDib::SetPalette(HPALETTE hpal)
{
	PALETTEENTRY *pal;
	int nColors = GetNumColors();

	// Check to see whether the DIB has a color table.
	if (nColors<=0) return;

	// Get a pointer to the RGB quads in the color table.
	RGBQUAD * pRGB = (RGBQUAD *) (lpDib + GetInfoHeaderSize());

	// Allocate a logical palette and fill it with the color table info.
	LOGPALETTE* pPal = (LOGPALETTE*) malloc(sizeof(LOGPALETTE) 
										+ nColors * sizeof(PALETTEENTRY));
	if (!pPal) return; // out of memory for logical palette
	pPal->palVersion = 0x300; // Windows 3.0
	pPal->palNumEntries = nColors; // Table size
	pal = pPal->palPalEntry;
	GetPaletteEntries(hpal,0,nColors,pal);
	for (int i=0; i<nColors; i++) {
		pRGB[i].rgbRed = pal[i].peRed;
		pRGB[i].rgbGreen = pal[i].peGreen;
		pRGB[i].rgbBlue = pal[i].peBlue;
	}
	free(pPal);
}

HPALETTE CDib::CreatePalette()
{
	HPALETTE hpal;
	int nColors = GetNumColors();

	// Check to see whether the DIB has a color table.
	if (nColors<=0) return 0;

	// Get a pointer to the RGB quads in the color table.
	RGBQUAD * pRGB = (RGBQUAD *) (lpDib + GetInfoHeaderSize());

	// Allocate a logical palette and fill it with the color table info.
	LOGPALETTE* pPal = (LOGPALETTE*) malloc(sizeof(LOGPALETTE) 
										+ nColors * sizeof(PALETTEENTRY));
	if (!pPal) return 0; // out of memory for logical palette
	pPal->palVersion = 0x300; // Windows 3.0
	pPal->palNumEntries = nColors; // Table size
	for (int i=0; i<nColors; i++) {
		pPal->palPalEntry[i].peRed = pRGB[i].rgbRed;
		pPal->palPalEntry[i].peGreen = pRGB[i].rgbGreen;
		pPal->palPalEntry[i].peBlue = pRGB[i].rgbBlue;
		pPal->palPalEntry[i].peFlags = 0;
	}
	hpal = ::CreatePalette(pPal);
	free(pPal);
	return hpal;
}

/*
* This routine converts bit count into a color count and vice versa
* They also verify that the bit count is one that is supported by windows,
* ie 1, 4, 8, and 24.  If the bit count is not supported, it returns -1
*/
inline int NColors(WORD bitCount)
{
  if (bitCount == 1 || bitCount == 4 || bitCount == 8)
    return 1 << bitCount;
  else if (bitCount == 24) return 0;
  return -1;
}
